/*********
https://www.baldengineer.com/arduino-lcd-display-tips.html
https://www.makerguides.com/character-i2c-lcd-arduino-tutorial/
https://deepbluembedded.com/lcd-custom-character-generator/
********
// https://circuitdigest.com/microcontroller-projects/littlefs-with-esp8266-to-read-write-and-delete-data-on-flash-memory-of-nodemcu
*/
//
//
#include "LiquidCrystal_I2C.h"
#include <LittleFS.h>

char FileName[] = __FILE__;
const char compile_date[] = __DATE__ " " __TIME__;

/////////////////////////////////////////////////////////////////////////////////////
//int CUnits = 60 * 1000;  //multiplier T1/T2 an TOn/Off for millis  60-> 1 min
int CUnits =  1000;  //multiplier T1/T2 an TOn/Off for millis  1s
//  see def  moded

//int CUnits = 300 * 1000;  //5 mins
//multiplier for timer  1 x1000 -> seconds, 60 -> minutes
//int CUnits = 60*60*1000  ; // hours  multiplier of millis()
/////////////////////////////////////////////////////////////////////////////////////

// Create lcd object of class LiquidCrystal_I2C:
LiquidCrystal_I2C lcd = LiquidCrystal_I2C(0x3f, 16, 2);  // Change to (0x27,20,4) for 20x4 LCD.
// LCD lines
char LCDline1[17];  // (0,0) 16 char + null
char LCDline2[17];  // (0,1) 16 char + null

long int TiMs;      // Timer: ms val when timer exhaust
long int OnTiMs ;   // cycle: ON duration start

long int cyclOn, cyclOff ;  // millis mark

//long int Tip = 0;  // ms last Ti
int T1 ;  // value preset
int T2 ;
int Tx ;  // last used timer  OR current active TOn or TOff param  T1 /TON  or T2 /TOff
int TOn ; // mode cycle
int TOff ;
int mode  ;     //Off On Tmr Cycl prefs  0 1 2 3 4  :  set default a pwr on

// default values T1 T2 see  CUnits
int Txd = 1;  //def sel to T1 and TOn
int T1d = 5 ;  //def val
int T2d = 30;
int TOnd= 5 ;  // TOnd
int TOffd=10 ;
int moded = 3 ;

int cytpct ;  // cycle percent

// mode  off  on  timer  cycli Prefs
//        0   1    2      3    4
char statusTxt[6];  //handlestatus timer" 5 + null

/////////////////////////////////////////////////////////////////////////////////////
// Make custom characters:
// char is cell  largeur x hauteur = 5 x 8  pixels    On OFF Timer
//uint8_t T10[] = {0x17, 0x15, 0x17, 0x00, 0x1d, 0x08, 0x09, 0x09}; //i o T

// file storing Tx
const char* filename = "/Tx.txt";
/////////////////////////////////////////////////////////////////////////////////////
// set pin numbers
const int buttonT1 = 0;   // the GPIO# of the pushbutton pin
const int buttonT2 = 2;   // the GPIO# of the pushbutton pin
const int buttonPl = 14;  // the GPIO# of the pushbutton pin
const int buttonMi = 12;  // the GPIO# of the pushbutton pin
const int buttonOn = 13;  // the GPIO# of the pushbutton pin

const int ledPin = 16;  // the GPIO# of the LED pin

#define LEDON HIGH
#define LEDOFF LOW

/////////////////////////////////////////////////////////////////////////////////////
void waitAfterTimer() {  //called at timer exhaust
  Serial.println("  waitAfterTimer()");
  int wl = 1;
  char LCDlineAlt[17];
  char waittxt[17] = "TmrOut,waiting..";  //will alttext with time
  //                  1234567890123456
  int waittxtdelay = 700;  //alt text delay
  long int MsZero = millis();

  sprintf(LCDline1, "         Off    ");  // mode or what else ?..
  //                1234567890123456
  lcd.setCursor(0, 1);  // second line
  lcd.print(LCDline1);

  sprintf(LCDlineAlt, "%s", waittxt);
  long int waitTgMs = millis();
  //  swap between two lines
  bool waitTg = 0;
  while (wl) {  //loop until button
    yield();
    if (!waitTg) {
      sprintf(LCDline1, "Wait%s %s", mconvertMillsToHHMMSS(millis() - MsZero), statusTxt);
      //                waitHH:MM:SS Off
      //                1234567890123456
      lcd.setCursor(0, 0);
      lcd.print(LCDline1);
    }
    if (!digitalRead(buttonOn) || !digitalRead(buttonT1) || !digitalRead(buttonT2)) {
      //loop button
      //pushed ..
      yield();
      lcd.setCursor(0, 0);
      lcd.print(LCDline1);
      Serial.println("buttonOn");
      wl = 0;
    } else {
      yield();
      if (millis() > waitTgMs + waittxtdelay) {  //after x sec ..
        lcd.setCursor(0, 0);
        lcd.print(LCDlineAlt);
        waitTgMs = millis();
        waitTg = !waitTg;
      }
    }
  }
}
/////////////////////////////////////////////////////////////////////////////////////
void setTiMs(int value = 0) {
  // value=0, get used timer Tx, or add/dec with passed value
  // return new TiMs
  long int tmil = millis();
  long int ret;
  //Serial.println() ;

  Serial.printf(">  setTiMs() : %i\n", value);
  Serial.printf(">  T1=%i  T2=%i\n", T1, T2);

  if (value == 0) {
    //Serial.println("> value==0");
    long int TiOffset;
    if (Tx == 1) {
      TiOffset = T1 * CUnits;
    } else {
      TiOffset = T2 * CUnits;
    }
    Serial.printf("> TiOffset : %i", TiOffset);
    Serial.println();
    tmil = millis();
    TiMs = tmil + TiOffset;
  } else {
    Serial.println();
    TiMs = tmil + (value * CUnits);
  }
  Serial.printf("> setTiMs exit: millis() = %i,  new TiMs = %i\n", tmil , TiMs);
}
/////////////////////////////////////////////////////////////////////////////////////
String mconvertMillsToHHMMSS(long int mils) {  // millis to  HH:MM:SS
  char ret[9];
  long int seconds = mils / 1000;
  long int hours = seconds / 3600;
  seconds %= 3600;
  long int minutes = seconds / 60;
  seconds %= 60;
  long int secs = seconds;
  sprintf(ret, "%02d:%02d.%02d", hours, minutes, seconds);  //8 chars
  return ret;
}
/////////////////////////////////////////////////////////////////////////////////////
void handlestatus() {
  // statusTxt is always  5 chars
  switch (mode) {
    case 0:                                       // OFF
      sprintf(statusTxt, "%s", String("Off  "));  //5 chars
      //                               12345
      OnTiMs = millis();
      digitalWrite(ledPin, LEDOFF);
      break;
    case 1:  // ON/OFF  and count duration
      sprintf(statusTxt, "%s", String("On   "));
      OnTiMs = millis();
      digitalWrite(ledPin, LEDON);
      break;
    case 2:  // timer
      sprintf(statusTxt, "%s", String("Timer"));
      setTiMs();
      break;
    case 3:  //cyclic power
      sprintf(statusTxt, "%s", String("Cycle")) ;
      OnTiMs = millis();
      break;
    case 4:  // settings
      sprintf(statusTxt, "%s", String("Prefs")) ;
      break;
  }
}
/////////////////////////////////////////////////////////////////////////////////////
void setTstop(){
	//           TOn   TOff
	//         -------      --------
	//         |      |     |
	//---------m      -------
	// millis  ^      ^      ^
	//                     cycOn
	//             cycOff
	long int mills = millis();
  cyclOff = mills + (TOn * CUnits) ;
  cyclOn = mills + ((TOn + TOff) * CUnits) ;
  Serial.printf("> setTstop Millis: %i, Off: %i, On: %i\n", mills , cyclOff, cyclOn ) ;
}
/////////////////////////////////////////////////////////////////////////////////////
void cyclepower() {
  //Serial.printf(" > 1. cyclepower \n");
  long int mills = millis();   //time flies ..
  //bool PwrOn= 1 ; //true ON
  //  cycOff  cycOn
  String OnOffs ;
  String HHMMSS ;
  String sel1, sel2   ;// active value TOn or TOff
  sel1=" "; sel2=" ";
  cytpct= TOn*100/(TOff)  ;
  if (Tx==1) {sel1=">";} else {sel2=">"; }  //  show active value
  if ( mills < cyclOff) { //running TOn
    sprintf(LCDline1, "%s%2iOn%s%3i", sel1, TOn,  mconvertMillsToHHMMSS(cyclOff - mills), cytpct ) ;
    //               1234567890123456
    //               >50Ofhh:mm.ssCyc
    //Serial.printf(" > 8. cyclepower LCDline2=!%s!\n" , LCDline2 );
    sprintf(LCDline2, "%s%2iOf%sCyc", sel2, TOff, "  :  .  " ) ;
  }
  else {  // in TOff
    sprintf(LCDline1, "%s%2iOn%s%3i", sel1, TOn, "  :  .  ", cytpct ) ;
    sprintf(LCDline2, "%s%2iOf%sCyc", sel2, TOff, mconvertMillsToHHMMSS(cyclOn - mills ) ) ;
    //                 1234567890123456
    //                 >10Onhh:mm.ss 16
   }
  if ( mills > cyclOff  ) {
    digitalWrite(ledPin, LEDOFF);
    }
  if ( mills > cyclOn  ) {
    digitalWrite(ledPin, LEDON);
    setTstop() ;
    }
}
/////////////////////////////////////////////////////////////////////////////////////
void reset2def(){
    T1 = T1d;
    T2 = T2d;
    Tx = Txd;
    TOn = TOnd ;
    TOff = TOffd ;
    mode=moded ;  handlestatus();
  }
/////////////////////////////////////////////////////////////////////////////////////
void preferences() {
  // T1 to sel pref   CUnits
  // + - to select option
  bool loopit=true ;
  int tt = CUnits;
  String unitset=".......";
  Serial.printf("> preferences\n" ) ;
  monparams() ;
  // current CUnits
  switch (CUnits) {
    case 1000:
      unitset="s" ;
      break ;
    case 60000:
      unitset="m" ;
      break ;
    case 3600000:
      unitset="h" ;
      break ;
    default:
      unitset="?" ;
    }

  //sprintf(LCDline1, "%s",  ".              .");
  sprintf(LCDline1,       "Pref (%s) On:exit", unitset);
  //                       Pref (m) On:exit
  //                       1234567890123456
  lcd.setCursor(0, 0); lcd.print(LCDline1);
  sprintf(LCDline2, "%s", "Red:units T1:Def");
  //                       1234567890123456
  lcd.setCursor(0, 1); lcd.print(LCDline2);
  while (loopit ) {
      yield();
      if (!digitalRead(buttonPl)) {
        //multiplier T1/T2 an TOn/Off for millis  1s
        delay(300) ;
        int tt;
        tt=CUnits ;
        switch (tt) {
          case 1000:
            CUnits = 60000; //minutes
            unitset="minutes" ;
            //       1234567
            break ;
          case 60000:
            CUnits = 60*60*1000; //3 600 000 msecs =hours
            unitset="hours  " ;
            //       1234567
            break ;
          case 3600000:
            CUnits =  1000; //seconds
            unitset="seconds" ;
            //       1234567
            break ;
          default:
            CUnits =  1000; //seconds
            unitset="seconds" ;
            //       1234567
        }
        sprintf(LCDline2, "Set max99%s     ", unitset ) ;
        //                 1234567890123456
        //                 Set max99minutes
        lcd.setCursor(0, 1); lcd.print(LCDline2);
        delay (500) ;
        monparams();
      }
    if (!digitalRead(buttonT1)) {  //  reset to def
      delay(300);
      sprintf(LCDline2, "SetDef?T1:y T2:n" ) ;
         //              1234567890123456
      lcd.setCursor(0, 1); lcd.print(LCDline2);
      bool loop2 =true ;
      while ( loop2  )  {
        yield();
        if (!digitalRead(buttonT1)) {
          reset2def();
          sprintf(LCDline2, "Done. Red or On " ) ;
          //                 1234567890123456
          lcd.setCursor(0, 1); lcd.print(LCDline2);
          delay (500);
          loop2=false;
        }
        if (!digitalRead(buttonT2)) {
          sprintf(LCDline2, "Red  or  Exit:On" ) ;
          //                 1234567890123456
          lcd.setCursor(0, 1); lcd.print(LCDline2);
          delay (500);
          loop2=false;
          }
      }
    }
    if (!digitalRead(buttonOn)) {
      delay(300) ;
      Serial.printf("> preferences exit\n" ) ;
      mode=0; handlestatus() ; //force to off
      setTiMs(); // if CUnits has changed
      Twrite();
      monparams() ;
      loopit =false ;
    }
  }  // loopit
  sprintf(LCDline1, "                ");
  lcd.setCursor(0, 0); lcd.print(LCDline1);
  sprintf(LCDline2, "                ");
  lcd.setCursor(0, 1); lcd.print(LCDline2);
  Twrite();
}
/////////////////////////////////////////////////////////////////////////////////////
void monparams() {
  Serial.printf("> monparams\n");
  Serial.printf(" CUnits: %i = %i secs\n", CUnits, CUnits/1000);
  Serial.printf(" Tx  : %i\n " , Tx );
  Serial.printf(" T1  : %i\n " , T1 );
  Serial.printf(" T2  : %i\n " , T2 );
  Serial.printf(" TOn : %i\n " , TOn );
  Serial.printf(" TOff: %i\n " , TOff );
  Serial.printf(" mode  : %i  %s\n", mode, statusTxt);
  Serial.printf("> ----------------------\n");
}
/////////////////////////////////////////////////////////////////////////////////////
int Twrite() {
  // Écrire les valeurs de T1 et T2 dans le fichier
  Serial.println(">  Twrite ..");
  File file = LittleFS.open(filename, "w");
  if (file) {
    file.println(T1, 0);  // (value, decimal places)
    file.println(T2, 0);
    file.println(Tx, 0);
    file.println(TOn, 0);
    file.println(TOff, 0);
    // if mode prefs (mode=4) store Off mode 0
    if (mode==4) { file.println(0, 0); handlestatus() ;}  else { file.println(mode, 0);  }
    file.println(CUnits, 0);
    file.close();
    Serial.println(">  Twrite exit  All params written.");
    //Serial.printf( "  T1:%i T2:%i Tx:%i\n", T1, T1, Tx);
    return true;
  } else {
    Serial.println(">  Twrite exit  Error file open.");
    return false;
  }
}
/////////////////////////////////////////////////////////////////////////////////////
void Tread() {
  Serial.println("> Read stored data:");
  if (LittleFS.exists(filename)) {
    File file = LittleFS.open(filename, "r");
    if (file) {
      String line = file.readStringUntil('\n');
      T1 = line.toInt();  line = file.readStringUntil('\n');
      T2 = line.toInt();  line = file.readStringUntil('\n');
      Tx = line.toInt();  line = file.readStringUntil('\n');
      TOn = line.toInt(); line = file.readStringUntil('\n');
      TOff = line.toInt();line = file.readStringUntil('\n');
      mode = line.toInt();line = file.readStringUntil('\n');
      CUnits = line.toInt(); line = file.readStringUntil('\n');

      file.close();
      handlestatus() ; //sync statustxt
    } else {
      Serial.println("  Erreur d'ouverture du fichier pour lecture");
    }
  } else {
    reset2def();
    Serial.println("  > File not found, set default values.");
    if (Twrite()) {
      Serial.println(" ok, Twrite() done");
    }
  }
}
/////////////////////////////////////////////////////////////////////////////////////
void setup() {
  Serial.begin(115200);  // start serial port
  delay(300);
  Serial.println("");
  Serial.println("");
  Serial.println("━━━━━━━━━━━━━━━━━━━━━━━━━━━━━");
  //              Compiled Nov  6 2023 19:00:57
// 17:08:18.957 -> Compiled Jan  1 2025 17:07:53
//                          1234567890123456

  Serial.println(__FILE__);
  Serial.println(String("Compiled ") + String(compile_date));
  Serial.println("━━━━━━━━━━━━━━━━━━━━━━━━━━━━━");
  //////////////////////////////////////////
  Serial.println(F("Read last params.."));
  if (!LittleFS.begin()) { Serial.println(" Erreur de montage LittleFS"); return; }
  Tread() ;
  //////////////////////////////////////////
  Serial.println(F("Set pins.."));
  //////////////////////////////////////////
  // initialize the pushbutton pin as an input
  pinMode(buttonT1, INPUT_PULLUP);
  pinMode(buttonT2, INPUT_PULLUP);
  pinMode(buttonPl, INPUT_PULLUP);
  pinMode(buttonMi, INPUT_PULLUP);
  pinMode(buttonOn, INPUT_PULLUP);
  // initialize the LED pin as an output
  pinMode(ledPin, OUTPUT);
  // power on at start
  digitalWrite(ledPin, LEDON);
  //////////////////////////////////////////
  Serial.println(F("Initialize LCD, show pgm name .."));
  //////////////////////////////////////////
  lcd.init();
  lcd.backlight();  //  noBacklight() ;
  lcd.clear();

  sprintf(LCDline1, "%s", FileName);
  sprintf(LCDline2, "%s", __DATE__);  // String(compile_date));
  lcd.setCursor(0, 0); lcd.print(LCDline1); // show first line  Serial.println(F("Initialize mode, TiMs, LCDclear, setTstop() .."));
  lcd.setCursor(0, 1); lcd.print(LCDline2); // show first line  Serial.println(F("Initialize mode, TiMs, LCDclear, setTstop() .."));
  //////////////////////////////////////////
  setTiMs();
  handlestatus();
  setTstop();
  delay(1500); // see the name  ?
  lcd.clear()  ;
  handlestatus();
  monparams() ;
  Serial.println(F("////// End Setup()"));
}
/////////////////////////////////////////////////////////////////////////////////////
void loop() {
  //To select a cursor location set the column 0..1 THEN the row 0..15
  //top left is 0,0
  // Like: lcd.setCursor(column, row)
  /*
	lcd.setCursor(0, 1);  // col 2, l 1
	lcd.write(1);  /ioT  char
	*/
	// cyclOn, cyclOff ;
  // Timer exhausted  ?
  String HHMMSS;  // will hold count/elapsed
  long int mills = millis();   //time flies ..
  String tpct = "      ";  //  pct of time elapsed   000pct
  //TiMs check
  //int switchstatus = mode;
  //Serial.printf(" 1. switch  %i : %s\n", mode , statusTxt);
  switch (mode) {
    /////////////
    case 0: //  On
    case 1: //  Off
      // show On or Off duration
      sprintf(LCDline1, "%s           ", statusTxt);
      //                 1234567890123456
      //                 12345
      HHMMSS = mconvertMillsToHHMMSS(mills - OnTiMs);
      sprintf(LCDline2, "%s        ", HHMMSS);
      break ;
    /////////////
    case 2: //  timer
      //if (mode == 2) { // when timer active
      //Serial.printf("> l 505 case 2\n");
      Serial.printf("> mills > TiMs ?  %i  %i\n",mills, TiMs);
      if (mills > TiMs) {  // if exhausted, wait button, show time OFF
        Serial.printf("> TRUE\n");
        mode = 0;
        handlestatus();
        Serial.printf(" TiMS = %i  exhausted, wait buttonOn", TiMs);
        //Serial.println(F(" exhausted, wait buttonOn"));
        digitalWrite(ledPin, LEDOFF);
        waitAfterTimer();  // loop until pushed
        digitalWrite(ledPin, LEDON);
        setTiMs(0);
        Serial.println(F(" timer reset"));
      }
      monparams();
      int T;  //hold on value for T1 T2
      if (Tx == 1) { T = T1; } else { T = T2;  }
      int ipct ;
      ipct = abs((TiMs - mills) * 100 / (T * CUnits) );
      tpct = String(ipct) + " pct  ";  // "  " to avoid double final't' when reaching <=9 pct, cast to 6 on sprintf
      Serial.printf(" TiMs:%i\n " , TiMs );
      Serial.printf(" mills:%i\n " , mills );
      Serial.printf(" T:%i\n " , T );
      Serial.printf(" CUnits:%i\n " , CUnits );
      Serial.printf(" ipct:%i\n " , ipct );
      Serial.printf(" tpct:%s\n " , tpct );
      // "99 pct" -> be 6 chars

      if (mode==2) {  //  only if timer
        //Serial.printf(">  l524 (mode==2) TRUE\n");
        if (Tx == 1) {  //set visual < >  indication of active timer value
          sprintf(LCDline1, ">%-3i<%-3i  %6s", T1, T2, tpct);
        } else {
          Serial.printf("try LCDline1:>%-3i<%-3i  %6s", T1, T2, tpct);
          sprintf(LCDline1, " %-3i>%-3i< %6s", T1, T2, tpct);
          //                 "1234567890123456"
          //                  >123<    .
          //                      >123< xxxpct
          //                      >123<
        }
      }
      ///////////////////////////////////////////////////////////////////
      HHMMSS = mconvertMillsToHHMMSS(TiMs - mills);
      sprintf(LCDline2, "%s   %s", HHMMSS, statusTxt);
      //                1234567890123456
      //                hh:mm.ss...Timer
      break;
    /////////////  cycle
    case 3:
      cyclepower() ;
      break;
  }  //  end switch

  // handle buttons /////////////////////////////////////////////////
  ///////////////////////////   T1  Ton
  if (!digitalRead(buttonT1)) {
      Serial.println("buttonT1");
      delay(300);
      //mode = 2;
      //handlestatus();
      Tx = 1;
      setTiMs(T1);
  }
  ///////////////////////////   T2  TOff
  if (!digitalRead(buttonT2)) {
    Serial.println("buttonT2");
    delay(300);
    //mode = 2;
    //handlestatus();
    Tx = 2;
    setTiMs(T2);

  }
  ///////////////////////////   + Plus
  if (!digitalRead(buttonPl)) {
    Serial.println("buttonPl");
    delay(300);
    switch(mode){
      ///////////////////////  	timer
      case 2:
        if (Tx == 1) {
          T1++;
          if (T1>99) {T1=99;} //only 2 digits :-(
        } else {
          T2++; if (T2>99) {T2=99;}
        }
        setTiMs(0);
        Twrite();
        break;
      ///////////////////////  cycle
      case 3:
        if (Tx == 1) { TOn++; if (TOn>99) {TOn=99;} }
        else { TOff++; if (TOff>99) {TOff=99;} }
        Twrite();
        setTstop();
    }
  }
  ///////////////////////////   - Minus
  if (!digitalRead(buttonMi)) {
    Serial.println("buttonMi");
    delay(300);
    switch(mode){
      case 2:////////////// timer
        if (Tx == 1) {
          T1--;
          if (T1 == 0) { T1 = 1;  }
        } else {
          T2--;
          if (T2 == 0) { T2 = 1;  }
        }
        setTiMs(0);
        Twrite();
        break;
      case 3:////////////// cycle
        if (Tx == 1) {
          TOn--;
          if (TOn == 0) { TOn = 1; }
        } else {
          TOff--;
          if (TOff == 0) {TOff = 1; }
        }
        Twrite();
        setTstop();
    }
}
  ///////////////////////////   Off On Tmr Cycl prefs
  //                            0   1   2  3    4
  if (!digitalRead(buttonOn)) {
    Serial.printf("> buttonOn\n");
    delay(300);
    mode++;
    if (mode == 5) {mode=0 ; } // last reached, loop
    handlestatus();
    monparams() ;
    switch (mode) {
      case 1:
      case 2:
        Serial.println("> On/Off: Reset timer");
        setTiMs(0);// reset timer to Tx val
        Serial.printf("> l 628 before break \n");
        break ;
      case 3:
        Serial.println("> cyclepower");
        cyclepower() ;
        break ;
      case 4:
        Serial.println("> preferences");
        preferences() ;
        break ;
    }
  }
  ///////////////////////////////////////////////////////////////////
  // Send to LCD  /////////////////////////////////////////////////
  lcd.setCursor(0, 0); lcd.print(LCDline1); // show first line
  lcd.setCursor(0, 1); lcd.print(LCDline2); // show second lin
  ///////////////////////////////////////////////////////////////////
}
/////////////////////////////////////////////////////////////////////////////////////
